#include "dictbrowser.h"

#include <QPalette>
#include <QScrollBar>
#include <QDebug>

DictBrowser::DictBrowser(QWidget *parent)
	: QTextBrowser(parent), histSize(100)
{
	plaintext = false;
	setOpenLinks(false);
	connect(this, SIGNAL(highlighted(const QString &)),
		SLOT(pointCursor(const QString &)));
	connect(this, SIGNAL(cursorPositionChanged()),
		SLOT(selection()));
}

void DictBrowser::moveTo(int steps)
{
	QList<HistoryEntry> &history = (steps > 0)
		? forwardHistory : backwardHistory;

	if (history.isEmpty())
		return;

	QList<HistoryEntry> &other = (steps > 0)
		? backwardHistory : forwardHistory;

	savePosition();
	if (steps < 0)
		steps = -steps;
	while (steps--) {
		other.prepend(entry);
		entry = history.takeFirst();
	}
	displayPage(entry.page);
	restorePosition();

	emit backwardAvailable(!backwardHistory.isEmpty());
	emit forwardAvailable(!forwardHistory.isEmpty());
	emit historyChanged(entry.title);
}

void DictBrowser::savePosition()
{
	entry.hpos = horizontalScrollBar()->value();
	entry.vpos = verticalScrollBar()->value();
}

void DictBrowser::restorePosition()
{
	horizontalScrollBar()->setValue(entry.hpos);
	verticalScrollBar()->setValue(entry.vpos);
}

void DictBrowser::clearHistory()
{
	backwardHistory.clear();
	forwardHistory.clear();
	emit backwardAvailable(false);
	emit forwardAvailable(false);
}

QStringList DictBrowser::historyList(const QList<HistoryEntry> &history, int len) const
{
	QStringList list;
	len = qMin(history.size(), len);
	for (int i = 0; i < len; ++i)
		list << history[i].title;
	return list;
}

void DictBrowser::setPage(const QString &query, const QString &text)
{
	savePosition();
	displayPage(text);

	if (entry.title != query && !entry.title.isEmpty())
		backwardHistory.prepend(entry);
	if (backwardHistory.size() > histSize)
		backwardHistory.removeLast();
	if (entry.title != query)
		forwardHistory.clear();
	entry.title = query;
	entry.page = text;

	emit backwardAvailable(!backwardHistory.isEmpty());
	emit forwardAvailable(!forwardHistory.isEmpty());
	emit historyChanged(query);
}

void DictBrowser::appendPage(const QString &query, const QString &text)
{
	savePosition();
	append(text);
	restorePosition();
	entry.page += text;
}

void DictBrowser::displayPage(const QString &text)
{
	if (plaintext) setPlainText(text);
	else setHtml(text);
}

QString DictBrowser::page()
{
	QString html = "<html><head><meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">\n";
	html += "<title>" + entry.title + "</title>";
	html += "\n<style type=\"text/css\">\n";
	html += document()->defaultStyleSheet();
	html += "\n</style></head><body>\n";
	html += entry.page;
	html += "\n</body></html>";
	return html;
}

void DictBrowser::plainText(bool toggle)
{
	plaintext = toggle;
	displayPage(entry.page);
}

void DictBrowser::setStyleSheet(const QString &style)
{
	document()->setDefaultStyleSheet(style);
}

void DictBrowser::setFont(const QFont &font)
{
	document()->setDefaultFont(font);
	displayPage(entry.page);
}

void DictBrowser::clearSelection()
{
	QTextCursor cursor = textCursor();
	cursor.clearSelection();
	setTextCursor(cursor);
}

void DictBrowser::busyCursor(bool busy)
{
	viewport()->setCursor((busy) ? Qt::BusyCursor : Qt::ArrowCursor);
	cursor = viewport()->cursor();
}

void DictBrowser::pointCursor(const QString &text)
{
	if (text.isEmpty())
		viewport()->setCursor(cursor);
}

void DictBrowser::focusOutEvent(QFocusEvent *event)
{
	QTextEdit::focusOutEvent(event);
}

void DictBrowser::selection()
{
	if (!textCursor().hasSelection())
		return;

	QPalette palette;
	QColor text = palette.color(QPalette::HighlightedText);
	QColor back = palette.color(QPalette::Highlight);

	if (textBackgroundColor() == back) {
		palette.setColor(QPalette::HighlightedText, back);
		palette.setColor(QPalette::Highlight, text);
	} else {
		palette.setColor(QPalette::HighlightedText, text);
		palette.setColor(QPalette::Highlight, back);
	}
	setPalette(palette);
}
