#include "dictpage.h"

#include <KLocale>
#include <QPalette>
#include <QDebug>

DictPage::DictPage(const DictResponse &response)
{
	set(response);
}

void DictPage::set(const DictResponse &response)
{
	switch (response.type) {
	case Words:
		words(response.text);
		break;
	case Matches:
		matches(response.matches);
		break;
	case DatabaseList:
		databases(response.list);
		break;
	case StrategyList:
		strategies(response.list);
		break;
	case DatabaseInfo:
		words(response.text);
		break;
	case ServerInfo:
		server(response.text);
		break;
	default:
		set(response.text);
	}
}

QString DictPage::htmltext()
{
	return html;
}

void DictPage::set(const QString &text)
{
	html = text;
}

void DictPage::nomatches(const QString &text, bool define, bool correct, bool suggestions)
{
	QString message = (define)
		? i18n("No definitions found for <b>%1</b>.", text)
		: i18n("No matches found for <b>%1</b>.", text);
	message = QString("<div class=\"message\">%1</div>").arg(message);

	if (suggestions)
		html.prepend(message + i18n("<p>Suggestions:"));
	else if (correct)
		set(message + i18n("<p>No suggestions."));
	else
		set(message);
}

void DictPage::heading(const QString &text)
{
	html.prepend("<div class=\"heading\">" + text + "</div>");
}

void DictPage::words(const QString &text)
{
	html = text;
	// replace html tags
	html.replace('<', "&lt;");
	html.replace('>', "&gt;");
	html.prepend("<div class=\"definition\">");
	
	int start = 0;
	while ((start = html.indexOf('{', start + 1)) >= 0) {
		int end = html.indexOf('}', start);
		// missing matching brace
		if (end < 0) break;

		int len = end - start;
		// too long for a link
		if (len > 240) continue;

		QString link = html.mid(start + 1, len - 1);
		// multiple lines inside braces, could be source code
		if (link.count('\n') > 1) continue;

		int pos;
		// skip external links
		if ((pos = link.indexOf('(')) >= 0 && link.indexOf(')') > pos) {
			html.replace(start, len + 1, link);
			continue;
		}
		// format link
		html.replace(start, len + 1, QString("<a href=\"%1\">%2</a>")
			.arg(link.simplified()).arg(link));
	}
	html.append("</div>");
}

void DictPage::server(const QString &text)
{
	html = "<div class=\"info\">";
	html += text;
	html += "</div>";
}

void DictPage::matches(const QStringList &list)
{
	html = "<div class=\"match\">";
	int count = qMin(list.size(), 240);
	for (int i = 0; i < count; ++i)
		html += QString("<a href=\"%1\">%1</a>\n").arg(list[i]);
	html += "</div>";
}

void DictPage::suggestions(const QString &database, const QStringList &list)
{
	html = "<div class=\"list\">" + database + ": ";
	html += "<span class=\"suggestion\">";
	int count = qMin(list.size(), 12);
	for (int i = 0; i < count; ++i)
		html += QString("<a href=\"%1\">%1</a>, ").arg(list[i]);
	html.chop(2);
	html += "</span>";
	html += "</div>";
}

void DictPage::databases(const DictList &list)
{
	html = "<table>";
	foreach (DictPair pair, list) {
		html += QString("<tr><td class=\"name\"><a href=\"about:%1\">%1</a><td>").arg(pair.first);
		html += "<td>" + pair.second;
	}
	html += "</table>";
}

void DictPage::strategies(const DictList &list)
{
	html = "<table>";
	foreach (DictPair pair, list) {
		html += "<tr><td class=\"name\">" + pair.first;
		html += "<td>" + pair.second;
	}
	html += "</table>";
}

void DictPage::error(const QString &text)
{
	html = "<div class=\"error\">";
	html += i18n("<h4>Error: %1</h4>", text);
}

QString DictPage::stylesheet()
{
	QPalette palette;
	QString style;
	style += ".definition { white-space: pre; margin-top: 1em }";
	style += ".match { white-space: pre }";
	style += QString(".heading { color: %1 }")
		.arg(palette.color(QPalette::HighlightedText).name());;
	style += QString(".heading { background: %1 }")
		.arg(palette.color(QPalette::Highlight).name());
	style += ".info { white-space: pre; font-family: monospace }";
	style += ".name { padding-right: 1em }";
	style += ".list { margin-top: 1em; margin-bottom: 1em }";
	style += ".message { margin-bottom: 1em }";
	style += "a { text-decoration: none }";
	style += QString("a { color: %1 }")
		.arg(palette.color(QPalette::Link).name());;
	return style;
}

void DictPage::about(const QString &text)
{
	html = "<div class=\"about\">";
	html += QString("<h3>%1</h3>").arg(text);
	html += "<p>Dikt is a dictionary that implements the client side of Dict protocol. Dikt is a network application, it looks up words on dict servers. You don't need to install any dictionaries, you only need a connection to Internet. You can find information about Dict protocol at <a href=\"http://www.dict.org/\">http://www.dict.org/</a>.";
	html += "<p>To look up a word, enter it into the query bar. By default Dikt searches <b>dict.org</b> dict server. You can add and edit hosts in Dikt settings and access them from the host menu. You can also switch to a different server by entering a URL in the form <b>dict://host/word</b>, where <b>host</b> is the dict server to connect to, and <b>word</b> is the word to look for.";
	html += "<p>A list of dict servers is available at <a href=\"http://luetzschena-stahmeln.de/dictd/\">http://luetzschena-stahmeln.de/dictd/</a>.";
	html += "<p>Dikt is very similar to web browsers, and it should be easy to use to anyone familiar with the Web and Internet. One option that needs some explanation is to yank selections. When it is set, Dikt will look up the text you select in your KDE and Qt applications.";
	html += "<p>For the latest release and for the contact information see <a href=\"http://diktv1.googlepages.com/\">Dikt web site</a>.";
	html += "<p></div>";
}

void DictPage::warning(const QString &text, Warning type)
{
	QString title, description;

	switch (type) {
	case LongSelection:
		title = i18n("Long Selection");
		description = i18n("Selected text is longer than configured limit");
		break;
	case UrlSelection:
		title = i18n("Url Selected");
		description = i18n("Selected text is a Url");
		break;
	}

	html = "<div class=\"warning\">";
	html += i18n("<h4>Warning: %1</h4>", title);
	html += QString("<p><a href=\"%1\">%1</a>").arg(text);
	html += QString("<p>%1.").arg(description);
	html += "</div>";
}
